#include "cppdefs.h"
      MODULE mod_mixing
!
!=======================================================================
!  Copyright (c) 2005 ROMS/TOMS Group                                  !
!================================================== Hernan G. Arango ===
!                                                                      !
!  Horizontal and vertical mixing coefficients:                        !
!                                                                      !
!  Akt          Vertical mixing coefficient (m2/s) for tracers.        !
!  dAktdz       Vertical gradient in mixing coefficient (m/s) for      !
!                 tracer 1, used in float random walk calculations     !
!  Akv          Vertical mixing coefficient (m2/s) for momentum.       !
!                                                                      !
!  Variables associated with the equation of state:                    !
!                                                                      !
!  alpha        Surface thermal expansion coefficient (1/Celsius).     !
!  beta         Surface saline contraction coefficient (1/PSU).        !
!  bvf          Brunt-Vaisala frequency squared (1/s2).                !
!  neutral      Coefficient to convert "in situ" density to neutral    !
!                 surface.                                             !
!                                                                      !
!  tke          Turbulent energy squared (m2/s2) at horizontal         !
!                 at W-points.                                         !
!  gls          Turbulent energy squared times turbulent length        !
!                 scale (m3/s2) at W-points.                           !
!                                                                      !
!  Large/McWilliams/Doney interior vertical mixing variables:          !
!                                                                      !
!  alfaobeta    Ratio of thermal expansion and saline contraction      !
!                 coefficients (Celsius/PSU) used in double            !
!                 diffusion.                                           !
!                                                                      !
!  Water clarity parameters:                                           !
!                                                                      !
!  Jwtype       Water clarity (Jerlov water type classification).      !
!                                                                      !
!  Large/McWilliams/Doney oceanic boundary layer variables:            !
!                                                                      !
!  ghats        Boundary layer nonlocal transport (T units/m).         !
!  hbbl         Depth of bottom oceanic boundary layer (m).            !
!  hsbl         Depth of surface oceanic boundary layer (m).           !
!  kbbl         Index of grid level above bottom  boundary layer.      !
!  ksbl         Index of grid level below surface boundary layer.      !
!                                                                      !
!=======================================================================
!
        USE mod_kinds

        implicit none

        TYPE T_MIXING
!
!  Nonlinear model state.
!
#ifdef SOLVE3D
# if defined LMD_SKPP || defined SOLAR_SOURCE
          integer,  pointer :: Jwtype(:,:)
# endif
# if defined LMD_SKPP || defined LMD_BKPP
          integer,  pointer :: ksbl(:,:)
#  ifdef LMD_BKPP
          integer,  pointer :: kbbl(:,:)
#  endif
# endif
#endif

#if defined UV_VIS2 || !defined SOLVE3D
          real(r8), pointer :: visc2_p(:,:)
          real(r8), pointer :: visc2_r(:,:)
#endif
#ifdef UV_VIS4
          real(r8), pointer :: visc4_p(:,:)
          real(r8), pointer :: visc4_r(:,:)
#endif
#ifdef SOLVE3D
# ifdef TS_DIF2
          real(r8), pointer :: diff2(:,:,:)
# endif
# ifdef TS_DIF4
          real(r8), pointer :: diff4(:,:,:)
# endif
          real(r8), pointer :: Akv(:,:,:)
          real(r8), pointer :: Akt(:,:,:,:)
# ifdef FLOAT_VWALK
          real(r8), pointer :: dAktdz(:,:,:)
# endif
# if defined LMD_SKPP || defined LMD_BKPP || defined BULK_FLUXES
          real(r8), pointer :: alpha(:,:)
          real(r8), pointer :: beta(:,:)
#  if defined TANGENT || defined TL_IOMS
          real(r8), pointer :: tl_alpha(:,:)
          real(r8), pointer :: tl_beta(:,:)
#  endif
#  ifdef ADJOINT
          real(r8), pointer :: ad_alpha(:,:)
          real(r8), pointer :: ad_beta(:,:)
#  endif
# endif
# ifdef BV_FREQUENCY
          real(r8), pointer :: bvf(:,:,:)
# endif
# if defined MIX_ISO_TS || defined GENT_McWILLIAMS
          real(r8), pointer :: neutral(:,:,:)
# endif
# if defined GLS_MIXING || defined MY25_MIXING
          real(r8), pointer :: tke(:,:,:,:)
          real(r8), pointer :: gls(:,:,:,:)
          real(r8), pointer :: Lscale(:,:,:)
          real(r8), pointer :: Akk(:,:,:)
#  ifdef GLS_MIXING
          real(r8), pointer :: Akp(:,:,:)
#  endif
# endif
# if defined LMD_MIXING && defined LMD_DDMIX
          real(r8), pointer :: alfaobeta(:,:,:)
# endif
# if defined LMD_SKPP || defined LMD_BKPP
          real(r8), pointer :: hsbl(:,:)
#  ifdef LMD_BKPP
          real(r8), pointer :: hbbl(:,:)
#  endif
#  ifdef LMD_NONLOCAL
          real(r8), pointer :: ghats(:,:,:,:)
#  endif
# endif
#endif

#if defined TANGENT || defined TL_IOMS
!
!  Tangent linear model state.
!
# ifdef SOLVE3D
          real(r8), pointer :: tl_Akv(:,:,:)
          real(r8), pointer :: tl_Akt(:,:,:,:)
# endif
#endif

#ifdef ADJOINT
!
!  Adjoint model state.
!
# ifdef SOLVE3D
          real(r8), pointer :: ad_Akv(:,:,:)
          real(r8), pointer :: ad_Akt(:,:,:,:)
# endif
#endif

#if defined FORWARD_READ && \
   (defined TANGENT || defined TL_IOMS || defined ADJOINT)
# ifdef FORWARD_MIXING
!
!  Latest two records of the nonlinear trajectory used to interpolate
!  the background state in the tangent linear and adjoint models.
!
          real(r8), pointer :: AkvG(:,:,:,:)
          real(r8), pointer :: AktG(:,:,:,:,:)
# endif
# if defined LMD_MIXING
          real(r8), pointer :: hsblG(:,:,:)
# endif
# if defined LMD_BKPP
          real(r8), pointer :: hbblG(:,:,:)
# endif
# if defined LMD_NONLOCAL
          real(r8), pointer :: ghatsG(:,:,:,:,:)
# endif
#endif

        END TYPE T_MIXING

        TYPE (T_MIXING), allocatable :: MIXING(:)

      CONTAINS
        
      SUBROUTINE allocate_mixing (ng, LBi, UBi, LBj, UBj)
!
!=======================================================================
!  Copyright (c) 2005 ROMS/TOMS Group                                  !
!================================================== Hernan G. Arango ===
!                                                                      !
!  This routine allocates all variables in the module for all nested   !
!  grids.                                                              !
!                                                                      !
!=======================================================================
!
      USE mod_param
!
!  Local variable declarations.
!
      integer, intent(in) :: ng, LBi, UBi, LBj, UBj
!
!-----------------------------------------------------------------------
!  Allocate module variables.
!-----------------------------------------------------------------------
!
      IF (ng.eq.1) allocate ( MIXING(Ngrids) )
! 
!  Nonlinear model state.
!
#if defined UV_VIS2 || !defined SOLVE3D
      allocate ( MIXING(ng) % visc2_p(LBi:UBi,LBj:UBj) )
      allocate ( MIXING(ng) % visc2_r(LBi:UBi,LBj:UBj) )
#endif

#ifdef UV_VIS4
      allocate ( MIXING(ng) % visc4_p(LBi:UBi,LBj:UBj) )
      allocate ( MIXING(ng) % visc4_r(LBi:UBi,LBj:UBj) )
#endif

#ifdef SOLVE3D
# ifdef TS_DIF2
      allocate ( MIXING(ng) % diff2(LBi:UBi,LBj:UBj,NT(ng)) )
# endif

# ifdef TS_DIF4
      allocate ( MIXING(ng) % diff4(LBi:UBi,LBj:UBj,NT(ng)) )
# endif

      allocate ( MIXING(ng) % Akv(LBi:UBi,LBj:UBj,0:N(ng)) )
      allocate ( MIXING(ng) % Akt(LBi:UBi,LBj:UBj,0:N(ng),NAT) )

# ifdef FLOAT_VWALK
      allocate ( MIXING(ng) % dAktdz(LBi:UBi,LBj:UBj,N(ng)) )
# endif

# if defined LMD_SKPP || defined LMD_BKPP || defined BULK_FLUXES
      allocate ( MIXING(ng) % alpha(LBi:UBi,LBj:UBj) )
      allocate ( MIXING(ng) % beta(LBi:UBi,LBj:UBj) )
# endif

# ifdef BV_FREQUENCY
      allocate ( MIXING(ng) % bvf(LBi:UBi,LBj:UBj,0:N(ng)) )
# endif

# if defined MIX_ISO_TS || defined GENT_McWILLIAMS
      allocate ( MIXING(ng) % neutral(LBi:UBi,LBj:UBj,0:N(ng)) )
# endif

# if defined GLS_MIXING || defined MY25_MIXING
      allocate ( MIXING(ng) % tke(LBi:UBi,LBj:UBj,0:N(ng),3) )
      allocate ( MIXING(ng) % gls(LBi:UBi,LBj:UBj,0:N(ng),3) )
      allocate ( MIXING(ng) % Lscale(LBi:UBi,LBj:UBj,0:N(ng)) )
      allocate ( MIXING(ng) % Akk(LBi:UBi,LBj:UBj,0:N(ng)) )

#  ifdef GLS_MIXING
      allocate ( MIXING(ng) % Akp(LBi:UBi,LBj:UBj,0:N(ng)) )
#  endif
# endif

# if defined LMD_MIXING && defined LMD_DDMIX
      allocate ( MIXING(ng) % alfaobeta(LBi:UBi,LBj:UBj,0:N(ng)) )
# endif

# if defined LMD_SKPP || defined SOLAR_SOURCE
      allocate ( MIXING(ng) % Jwtype(LBi:UBi,LBj:UBj) )
# endif

# if defined LMD_SKPP || defined LMD_BKPP
      allocate ( MIXING(ng) % ksbl(LBi:UBi,LBj:UBj) )
      allocate ( MIXING(ng) % hsbl(LBi:UBi,LBj:UBj) )

#  ifdef LMD_BKPP
      allocate ( MIXING(ng) % kbbl(LBi:UBi,LBj:UBj) )
      allocate ( MIXING(ng) % hbbl(LBi:UBi,LBj:UBj) )
#  endif

#  ifdef LMD_NONLOCAL
      allocate ( MIXING(ng) % ghats(LBi:UBi,LBj:UBj,0:N(ng),NAT) )
#  endif

# endif
#endif

#if defined TANGENT || defined TL_IOMS
!
!  Tangent linear model state.
!
# ifdef SOLVE3D
      allocate ( MIXING(ng) % tl_Akv(LBi:UBi,LBj:UBj,0:N(ng)) )
      allocate ( MIXING(ng) % tl_Akt(LBi:UBi,LBj:UBj,0:N(ng),NAT) )
# endif

# if defined LMD_SKPP || defined LMD_BKPP || defined BULK_FLUXES
      allocate ( MIXING(ng) % tl_alpha(LBi:UBi,LBj:UBj) )
      allocate ( MIXING(ng) % tl_beta(LBi:UBi,LBj:UBj) )
# endif

#endif

#ifdef ADJOINT
!
!  Adjoint model state.
!
# ifdef SOLVE3D
      allocate ( MIXING(ng) % ad_Akv(LBi:UBi,LBj:UBj,0:N(ng)) )
      allocate ( MIXING(ng) % ad_Akt(LBi:UBi,LBj:UBj,0:N(ng),NAT) )
# endif
# if defined LMD_SKPP || defined LMD_BKPP || defined BULK_FLUXES
      allocate ( MIXING(ng) % ad_alpha(LBi:UBi,LBj:UBj) )
      allocate ( MIXING(ng) % ad_beta(LBi:UBi,LBj:UBj) )
# endif
#endif

#if defined FORWARD_READ && \
   (defined TANGENT || defined TL_IOMS || defined ADJOINT)
# ifdef FORWARD_MIXING
!
!  Latest two records of the nonlinear trajectory used to interpolate
!  the background state in the tangent linear and adjoint models.
!
      allocate ( MIXING(ng) % AkvG(LBi:UBi,LBj:UBj,0:N(ng),2) )
      allocate ( MIXING(ng) % AktG(LBi:UBi,LBj:UBj,0:N(ng),2,NAT) )
# endif
# if defined LMD_MIXING
      allocate ( MIXING(ng) % hsblG(LBi:UBi,LBj:UBj,2) )
# endif
# if defined LMD_BKPP
      allocate ( MIXING(ng) % hbblG(LBi:UBi,LBj:UBj,2) )
# endif
# if defined LMD_NONLOCAL
      allocate ( MIXING(ng) % ghatsG(LBi:UBi,LBj:UBj,0:N(ng),2,NAT) )
# endif
#endif

      RETURN
      END SUBROUTINE allocate_mixing

      SUBROUTINE initialize_mixing (ng, tile)
!
!=======================================================================
!  Copyright (c) 2005 ROMS/TOMS Group                                  !
!================================================== Hernan G. Arango ===
!                                                                      !
!  This routine allocates and initializes all variables in module      !
!  "mod_mixing" for all nested grids.                                  !
!                                                                      !
!=======================================================================
!
      USE mod_param
      USE mod_scalars
!
!  Imported variable declarations.
!
      integer, intent(in) :: ng, tile
!
!  Local variable declarations.
!
      integer :: Imin, Imax, Jmin, Jmax
      integer :: i, j
#ifdef SOLVE3D
      integer :: itrc, k
#endif

      real(r8), parameter :: IniVal = 0.0_r8
!
#include "tile.h"
!
!  Set array initialization range.
!
#ifdef _OPENMP
      IF (WESTERN_EDGE) THEN
        Imin=LBi
      ELSE
        Imin=Istr
      END IF
      IF (EASTERN_EDGE) THEN
        Imax=UBi
      ELSE
        Imax=Iend
      END IF
      IF (SOUTHERN_EDGE) THEN
        Jmin=LBj
      ELSE
        Jmin=Jstr
      END IF
      IF (NORTHERN_EDGE) THEN
        Jmax=UBj
      ELSE
        Jmax=Jend
      END IF
#else
      Imin=LBi
      Imax=UBi
      Jmin=LBj
      Jmax=UBj
#endif
!
!-----------------------------------------------------------------------
!  Initialize module variables.
!-----------------------------------------------------------------------
!
!  Nonlinear model state.
!
      DO j=Jmin,Jmax
#if defined UV_VIS2 || !defined SOLVE3D
        DO i=Imin,Imax
          MIXING(ng) % visc2_p(i,j) = visc2(ng)
          MIXING(ng) % visc2_r(i,j) = visc2(ng)
        END DO
#endif
#ifdef UV_VIS4
        DO i=Imin,Imax
          MIXING(ng) % visc4_p(i,j) = visc4(ng)
          MIXING(ng) % visc4_r(i,j) = visc4(ng)
        END DO
#endif
#ifdef SOLVE3D
# ifdef TS_DIF2
        DO itrc=1,NT(ng)
          DO i=Imin,Imax
            MIXING(ng) % diff2(i,j,itrc) = tnu2(itrc,ng)
          END DO
        END DO
# endif
# ifdef TS_DIF4
        DO itrc=1,NT(ng)
          DO i=Imin,Imax
            MIXING(ng) % diff4(i,j,itrc) = tnu4(itrc,ng)
          END DO
        END DO
# endif
        DO i=Imin,Imax
          MIXING(ng) % Akv(i,j,0) = IniVal
          MIXING(ng) % Akv(i,j,N(ng)) = IniVal
        END DO
        DO k=1,N(ng)-1
          DO i=Imin,Imax
            MIXING(ng) % Akv(i,j,k) = Akv_bak(ng)
          END DO
        END DO
        DO itrc=1,NAT
          DO i=Imin,Imax
            MIXING(ng) % Akt(i,j,0,itrc) = IniVal
            MIXING(ng) % Akt(i,j,N(ng),itrc) = IniVal
          END DO
          DO k=1,N(ng)-1
            DO i=Imin,Imax
              MIXING(ng) % Akt(i,j,k,itrc) = Akt_bak(itrc,ng)
            END DO
          END DO
        END DO
# ifdef FLOAT_VWALK
        DO k=1,N(ng)
          DO i=Imin,Imax
            MIXING(ng) % dAktdz(i,j,k) = IniVal
          END DO
        END DO
# endif
# if defined LMD_SKPP || defined LMD_BKPP || defined BULK_FLUXES
        DO i=Imin,Imax
          MIXING(ng) % alpha(i,j) = IniVal
          MIXING(ng) % beta(i,j) = IniVal
        END DO
# endif
# ifdef BV_FREQUENCY
        DO k=0,N(ng)
          DO i=Imin,Imax
            MIXING(ng) % bvf(i,j,k) = IniVal
          END DO
        END DO
# endif
# if defined MIX_ISO_TS || defined GENT_McWILLIAMS
        DO k=1,N(ng)
          DO i=Imin,Imax
            MIXING(ng) % neutral(i,j,k) = IniVal
          END DO
        END DO
# endif

# if defined GLS_MIXING || defined MY25_MIXING
        DO k=0,N(ng)
          DO i=Imin,Imax
            MIXING(ng) % tke(i,j,k,1) = gls_Kmin(ng)
            MIXING(ng) % tke(i,j,k,2) = gls_Kmin(ng)
            MIXING(ng) % tke(i,j,k,3) = gls_Kmin(ng)
            MIXING(ng) % gls(i,j,k,1) = gls_Pmin(ng)
            MIXING(ng) % gls(i,j,k,2) = gls_Pmin(ng)
            MIXING(ng) % gls(i,j,k,3) = gls_Pmin(ng)
            MIXING(ng) % Lscale(i,j,k) = IniVal
          END DO
        END DO
        DO i=Imin,Imax
          MIXING(ng) % Akk(i,j,0) = IniVal
          MIXING(ng) % Akk(i,j,N(ng)) = IniVal
#  ifdef GLS_MIXING
          MIXING(ng) % Akp(i,j,0) = IniVal
          MIXING(ng) % Akp(i,j,N(ng)) = IniVal
#  endif
        END DO
        DO k=1,N(ng)-1
          DO i=Imin,Imax
            MIXING(ng) % Akk(i,j,k) = Akk_bak(ng)
#  ifdef GLS_MIXING
            MIXING(ng) % Akp(i,j,k) = Akk_bak(ng)
#  endif
          END DO
        END DO
# endif
# if defined LMD_MIXING && defined LMD_DDMIX
        DO k=0,N(ng)
          DO i=Imin,Imax
            MIXING(ng) % alfaobeta(i,j,k) = IniVal
          END DO
        END DO
# endif
# if defined LMD_SKPP || defined SOLAR_SOURCE
        DO i=Imin,Imax
          MIXING(ng) % Jwtype(i,j) = lmd_Jwt(ng)
        END DO
# endif
# if defined LMD_SKPP || defined LMD_BKPP
        DO i=Imin,Imax
          MIXING(ng) % ksbl(i,j) = 0
          MIXING(ng) % hsbl(i,j) = IniVal
        END DO
#  ifdef LMD_BKPP
        DO i=Imin,Imax
          MIXING(ng) % kbbl(i,j) = 0
          MIXING(ng) % hbbl(i,j) = IniVal
        END DO
#  endif
#  ifdef LMD_NONLOCAL
        DO itrc=1,NAT
          DO k=0,N(ng)
            DO i=Imin,Imax
              MIXING(ng) % ghats(i,j,k,itrc) = IniVal
            END DO
          END DO
        END DO
#  endif
# endif
#endif
      END DO

#if defined TANGENT || defined TL_IOMS
!
!  Tangent linear model state.
!
# ifdef SOLVE3D
      DO j=Jmin,Jmax
        DO k=0,N(ng)           
          DO i=Imin,Imax
            MIXING(ng) % tl_Akv(i,j,k) = IniVal
          END DO
        END DO
        DO itrc=1,NAT
          DO k=0,N(ng)
            DO i=Imin,Imax
              MIXING(ng) % tl_Akt(i,j,k,itrc) = IniVal
            END DO
          END DO
        END DO
#  if defined LMD_SKPP || defined LMD_BKPP || defined BULK_FLUXES
        DO i=Imin,Imax
          MIXING(ng) % tl_alpha(i,j) = IniVal
          MIXING(ng) % tl_beta(i,j) = IniVal
        END DO
#  endif
      END DO
# endif
#endif

#ifdef ADJOINT
!
!  Adjoint model state.
!
# ifdef SOLVE3D
      DO j=Jmin,Jmax
        DO k=0,N(ng)           
          DO i=Imin,Imax
            MIXING(ng) % ad_Akv(i,j,k) = IniVal
          END DO
        END DO
        DO itrc=1,NAT
          DO k=0,N(ng)
            DO i=Imin,Imax
              MIXING(ng) % ad_Akt(i,j,k,itrc) = IniVal
            END DO
          END DO
        END DO
#  if defined LMD_SKPP || defined LMD_BKPP || defined BULK_FLUXES
        DO i=Imin,Imax
          MIXING(ng) % ad_alpha(i,j) = IniVal
          MIXING(ng) % ad_beta(i,j) = IniVal
        END DO
#  endif
      END DO
# endif
#endif

#if defined FORWARD_READ && defined FOWARD_MIXING && defined SOLVE3D && \
   (defined TANGENT || defined TL_IOMS || defined ADJOINT)
!
!  Latest two records of the nonlinear trajectory used to interpolate
!  the background state in the tangent linear and adjoint models.
!
      DO j=Jmin,Jmax
# ifdef FORWARD_MIXING
        DO k=0,N(ng)
          DO i=Imin,Imax
            MIXING(ng) % AkvG(i,j,k,1) = IniVal
            MIXING(ng) % AkvG(i,j,k,2) = IniVal
          END DO
        END DO
        DO itrc=1,NAT
          DO k=0,N(ng)
            DO i=Imin,Imax
              MIXING(ng) % AktG(i,j,k,1,itrc) = IniVal
              MIXING(ng) % AktG(i,j,k,2,itrc) = IniVal
            END DO
          END DO
        END DO
# endif
# if defined LMD_MIXING
        DO i=Imin,Imax
          MIXING(ng) % hsblG(i,j,1) = IniVal
          MIXING(ng) % hsblG(i,j,2) = IniVal
        END DO
# endif
# if defined LMD_BKPP
        DO i=Imin,Imax
          MIXING(ng) % hbblG(i,j,1) = IniVal
          MIXING(ng) % hbblG(i,j,2) = IniVal
        END DO
# endif
# if defined LMD_NONLOCAL
        DO itrc=1,NAT
          DO k=0,N(ng)
            DO i=Imin,Imax
              MIXING(ng) % ghatsG(i,j,0:N(ng),1,itrc) = IniVal
              MIXING(ng) % ghatsG(i,j,0:N(ng),2,itrc) = IniVal
            END DO
          END DO
        END DO
# endif
      END DO
#endif

      RETURN
      END SUBROUTINE initialize_mixing

      END MODULE mod_mixing
