      MODULE convolution_mod

#ifdef IS4DVAR
!
!=================================================== Andrew M. Moore ===
!  Copyright (c) ROMS/TOMS Adjoint Group                               !
!================================================== Hernan G. Arango ===
!                                                                      !
!  In incremental 4DVAR, the minimization conditioning is dominated    !
!  by the background error covariance matrix, B,  so a new variable    !
!  is defined to transform the the state vector, X,                    !
!                                                                      !
!              V = B^(-1/2) X            (V-space)                     !
!                                                                      !
!  and the cost function gradient becomes                              !
!                                                                      !
!          dJ/dV = transpose[B^(1/2)] dJ/dX                            !
!                                                                      !
!  where  dJ/dX is computed by integrating the adjoint model forced    !
!  by the misfit between model and observations.  The  minimization    !
!  of J in the descent algorithm is done in V-space.                   !
!                                                                      !
!  Usually,  a generalized diffusion equation  is used for defining    !
!  correlation models for B.  These spatial convolutions constitute    !
!  a smoothing action on the correlation operator  and  are used to    !
!  transform between X-space to V-space and vice versa:                !
!                                                                      !
!    ad_convolution     compute dJ/dV from dJ/dX                       !
!    tl_convolution     compute X from V                               !
!                                                                      !
!======================================================================!
!
      implicit none
!
      PRIVATE
      PUBLIC :: ad_convolution
      PUBLIC :: tl_convolution

      CONTAINS

      SUBROUTINE ad_convolution (ng, tile, Linp)
!
!=======================================================================
!                                                                      !
!  This routine computes adjoint spatial convolution of requested      !
!  adjoint state solution.                                             !
!                                                                      !
!=======================================================================
!
      USE mod_param
# ifdef SOLVE3D
      USE mod_coupling
# endif
      USE mod_grid
      USE mod_ocean
!
!  Imported variable declarations.
!
      integer, intent(in) :: ng, tile, Linp
!
!  Local variable declarations.
!
# include "tile.h"
!
# ifdef PROFILE
      CALL wclock_on (ng, iADM, ?)
# endif
      CALL ad_convolution_tile (ng, Istr, Iend, Jstr, Jend,             &
     &                          LBi, UBi, LBj, UBj,                     &
     &                          Linp,                                   &
     &                          GRID(ng) % pm,                          &
     &                          GRID(ng) % pmon_r,                      &
     &                          GRID(ng) % pmon_u,                      &
     &                          GRID(ng) % pn,                          &
     &                          GRID(ng) % pnom_p,                      &
     &                          GRID(ng) % pnom_v,                      &
# ifdef MASKING
     &                          GRID(ng) % pmask,                       &
     &                          GRID(ng) % rmask,                       &
     &                          GRID(ng) % umask,                       &
     &                          GRID(ng) % vmask,                       &
# endif
# ifdef SOLVE3D
     &                          GRID(ng) % h,                           &
     &                          GRID(ng) % Hz,                          &
     &                          GRID(ng) % z_r,                         &
     &                          GRID(ng) % z_w,                         &
#  ifdef ICESHELF
     &                          GRID(ng) % zice,                        &
#  endif
     &                          COUPLING(ng) % Zt_avg1,                 &
#  if defined SEdIMENT && defined SED_MORPH
     &                          OCEAN(ng) % bed,                        &
     &                          GRID(ng) % bed_thick0,                  &
#  endif
# endif
# ifdef SOLVE3D
     &                          OCEAN(ng) % b_u,                        &
     &                          OCEAN(ng) % b_v,                        &
     &                          OCEAN(ng) % b_t,                        &
# else
     &                          OCEAN(ng) % b_ubar,                     &
     &                          OCEAN(ng) % b_vbar,                     &
# endif
     &                          OCEAN(ng) % b_zeta,                     &
# ifdef SOLVE3D
     &                          OCEAN(ng) % ad_t,                       &
     &                          OCEAN(ng) % ad_u,                       &
     &                          OCEAN(ng) % ad_v,                       &
# else
     &                          OCEAN(ng) % ad_ubar,                    &
     &                          OCEAN(ng) % ad_vbar,                    &
# endif
     &                          OCEAN(ng) % ad_zeta)
# ifdef PROFILE
      CALL wclock_off (ng, iADM, ?)
# endif
      RETURN
      END SUBROUTINE ad_convolution
!
!***********************************************************************
      SUBROUTINE ad_convolution_tile (ng, Istr, Iend, Jstr, Jend,       &
     &                                LBi, UBi, LBj, UBj,               &
     &                                Linp,                             &
     &                                pm, pmon_r, pmon_u,               &
     &                                pn, pnom_p, pnom_v,               &
# ifdef MASKING
     &                                pmask, rmask, umask, vmask,       &
# endif
# ifdef SOLVE3D
     &                                h, Hz,                            &
     &                                z_r, z_w,                         &
#  ifdef ICESHELF
     &                                zice,                             &
#  endif
     &                                Zt_avg1,                          &
#  if defined SEdIMENT && defined SED_MORPH
     &                                bed, bed_thick0,                  &
#  endif
# endif
# ifdef SOLVE3D
     &                                b_u, b_v, b_t,                    &
# else
     &                                b_ubar, b_vbar,                   &
# endif
     &                                b_zeta,                           &
# ifdef SOLVE3D
     &                                ad_u, ad_v, ad_t,                 &
# else
     &                                ad_ubar, ad_vbar,                 &
# endif
     &                                ad_zeta)
!***********************************************************************
!
      USE mod_param
      USE mod_fourdvar
      USE mod_scalars
!
      USE ad_bcov_2d_mod
# ifdef SOLVE3D
      USE ad_bcov_3d_mod
      USE set_depth_mod, ONLY : set_depth_tile
# endif
!
!  Imported variable declarations.
!
      integer, intent(in) :: ng, Iend, Istr, Jend, Jstr
      integer, intent(in) :: LBi, UBi, LBj, UBj
      integer, intent(in) :: Linp
!
# ifdef ASSUMED_SHAPE
      real(r8), intent(in) :: pm(LBi:,LBj:)
      real(r8), intent(in) :: pmon_r(LBi:,LBj:)
      real(r8), intent(in) :: pmon_u(LBi:,LBj:)
      real(r8), intent(in) :: pn(LBi:,LBj:)
      real(r8), intent(in) :: pnom_p(LBi:,LBj:)
      real(r8), intent(in) :: pnom_v(LBi:,LBj:)
#  ifdef MASKING
      real(r8), intent(in) :: pmask(LBi:,LBj:)
      real(r8), intent(in) :: rmask(LBi:,LBj:)
      real(r8), intent(in) :: umask(LBi:,LBj:)
      real(r8), intent(in) :: vmask(LBi:,LBj:)
#  endif
#  ifdef SOLVE3D
      real(r8), intent(in) :: b_u(LBi:,LBj:,:)
      real(r8), intent(in) :: b_v(LBi:,LBj:,:)
      real(r8), intent(in) :: b_t(LBi:,LBj:,:,:)
#  else
      real(r8), intent(in) :: b_ubar(LBi:,LBj:)
      real(r8), intent(in) :: b_vbar(LBi:,LBj:)
#  endif
      real(r8), intent(in) :: b_zeta(LBi:,LBj:)
#  ifdef SOLVE3D
#   ifdef ICESHELF
      real(r8), intent(in) :: zice(LBi:,LBj:)
#   endif
#   if defined SEdIMENT && defined SED_MORPH
      real(r8), intent(in) :: bed(LBi:,LBj:,:,:)

      real(r8), intent(inout) :: bed_thick0(LBi:,LBj:)
#   endif
      real(r8), intent(inout) :: h(LBi:,LBj:)
      real(r8), intent(inout) :: Hz(LBi:,LBj:,:)
      real(r8), intent(inout) :: Zt_avg1(LBi:,LBj:)
      real(r8), intent(inout) :: z_r(LBi:,LBj:,:)
      real(r8), intent(inout) :: z_w(LBi:,LBj:,:)
#  endif
#  ifdef SOLVE3D
      real(r8), intent(inout) :: ad_u(LBi:,LBj:,:,:)
      real(r8), intent(inout) :: ad_v(LBi:,LBj:,:,:)
      real(r8), intent(inout) :: ad_t(LBi:,LBj:,:,:,:)
#  else
      real(r8), intent(inout) :: ad_ubar(LBi:,LBj:,:)
      real(r8), intent(inout) :: ad_vbar(LBi:,LBj:,:)
#  endif
      real(r8), intent(inout) :: ad_zeta(LBi:,LBj:,:)

# else

      real(r8), intent(in) :: pm(LBi:UBi,LBj:UBj)
      real(r8), intent(in) :: pmon_r(LBi:UBi,LBj:UBj)
      real(r8), intent(in) :: pmon_u(LBi:UBi,LBj:UBj)
      real(r8), intent(in) :: pn(LBi:UBi,LBj:UBj)
      real(r8), intent(in) :: pnom_p(LBi:UBi,LBj:UBj)
      real(r8), intent(in) :: pnom_v(LBi:UBi,LBj:UBj)
#  ifdef MASKING
      real(r8), intent(in) :: pmask(LBi:UBi,LBj:UBj)
      real(r8), intent(in) :: rmask(LBi:UBi,LBj:UBj)
      real(r8), intent(in) :: umask(LBi:UBi,LBj:UBj)
      real(r8), intent(in) :: vmask(LBi:UBi,LBj:UBj)
#  endif
#  ifdef SOLVE3D
      real(r8), intent(in) :: b_u(LBi:UBi,LBj:UBj,N(ng))
      real(r8), intent(in) :: b_v(LBi:UBi,LBj:UBj,N(ng))
      real(r8), intent(in) :: b_t(LBi:UBi,LBj:UBj,N(ng),NT(ng))
#  else
      real(r8), intent(in) :: b_ubar(LBi:UBi,LBj:UBj)
      real(r8), intent(in) :: b_vbar(LBi:UBi,LBj:UBj)
#  endif
      real(r8), intent(in) :: b_zeta(LBi:UBi,LBj:UBj)
#  ifdef SOLVE3D
#   ifdef ICESHELF
      real(r8), intent(in) :: zice(LBi:UBi,LBj:UBj)
#   endif
#   if defined SEdIMENT && defined SED_MORPH
      real(r8), intent(in) :: bed(LBi:UBi,LBj:UBj,Nbed,MBEDP)

      real(r8), intent(inout) :: bed_thick0(LBi:UBi,LBj:UBj)
#   endif
      real(r8), intent(inout) :: h(LBi:UBi,LBj:UBj)
      real(r8), intent(inout) :: Hz(LBi:UBi,LBj:UBj,N(ng))
      real(r8), intent(inout) :: Zt_avg1(LBi:UBi,LBj:UBj)
      real(r8), intent(inout) :: z_r(LBi:UBi,LBj:UBj,N(ng))
      real(r8), intent(inout) :: z_w(LBi:UBi,LBj:UBj,0:N(ng))
#  endif
#  ifdef SOLVE3D
      real(r8), intent(inout) :: ad_u(LBi:UBi,LBj:UBj,N(ng),2)
      real(r8), intent(inout) :: ad_v(LBi:UBi,LBj:UBj,N(ng),2)
      real(r8), intent(inout) :: ad_t(LBi:UBi,LBj:UBj,N(ng),3,NT(ng))
#  else
      real(r8), intent(inout) :: ad_ubar(LBi:UBi,LBj:UBj,3)
      real(r8), intent(inout) :: ad_vbar(LBi:UBi,LBj:UBj,3)
#  endif
      real(r8), intent(inout) :: ad_zeta(LBi:UBi,LBj:UBj,3)
# endif
!
!  Local variable declarations.
!
      integer :: IstrR, IendR, JstrR, JendR, IstrU, JstrV
      integer :: i, itrc, j
!
# include "set_bounds.h"

# ifdef SOLVE3D
!
!-----------------------------------------------------------------------
!  Compute depths and thicknesses for convolution. Use a rest state
!  free-surface.
!-----------------------------------------------------------------------
!
      DO j=JstrR,JendR
        DO i=IstrR,IendR
          Zt_avg1(i,j)=0.0_r8
        END DO
      END DO
      CALL set_depth_tile (ng, Istr, Iend, Jstr, Jend,                  &
     &                     LBi, UBi, LBj, UBj,                          &
     &                     h,                                           &
#  ifdef ICESHELF
     &                     zice,                                        &
#  endif
#  if defined SEDIMENT && defined SED_MORPH
     &                     bed, bed_thick0,                             &
#  endif
     &                     Zt_avg1,                                     &
     &                     Hz, z_r, z_w)
# endif
!
!-----------------------------------------------------------------------
!  Adjoint spatial convolutions.
!-----------------------------------------------------------------------
!
!  Free-surface.
!
      CALL ad_bcov_r2d_tile (ng, iADM, Istr, Iend, Jstr, Jend,          &
     &                       LBi, UBi, LBj, UBj, Nghost,                &
     &                       DTHdif, Hdecay,                            &
     &                       b_zeta,                                    &
     &                       pm, pn, pmon_u, pnom_v,                    &
# ifdef MASKING
     &                       umask, vmask,                              &
# endif
     &                       ad_zeta(:,:,Linp))

# ifndef SOLVE3D
!
!  2D U-velocity.
!
      CALL ad_bcov_u2d_tile (ng, iADM, Istr, Iend, Jstr, Jend,          &
     &                       LBi, UBi, LBj, UBj, Nghost,                &
     &                       DTHdif, Hdecay,                            &
     &                       b_ubar,                                    &
     &                       pm, pn, pmon_r, pnom_p,                    &
#  ifdef MASKING
     &                       pmask,                                     &
#  endif
     &                       ad_ubar(:,:,Linp))
!
!  2D V-velocity.
!
      CALL ad_bcov_v2d_tile (ng, iADM, Istr, Iend, Jstr, Jend,          &
     &                       LBi, UBi, LBj, UBj, Nghost,                &
     &                       DTHdif, Hdecay,                            &
     &                       b_vbar,                                    &
     &                       pm, pn, pmon_p, pnom_r,                    &
#  ifdef MASKING
     &                       pmask,                                     &
#  endif
     &                       ad_vbar(:,:,Linp))
# else
!
!  3D U-velocity.
!
      CALL ad_bcov_u3d_tile (ng, iADM, Istr, Iend, Jstr, Jend,          &
     &                       LBi, UBi, LBj, UBj, 1, N(ng), Nghost,      &
     &                       DTHdif, DTVdif, Hdecay, Vdecay,            &
     &                       b_u,                                       &
     &                       pm, pn, pmon_r, pnom_p,                    &
#  ifdef MASKING
     &                       pmask,                                     &
#  endif
     &                       Hz, z_r,                                   &
     &                       ad_u(:,:,:,Linp))
!
!  3D V-velocity.
!
      CALL ad_bcov_v3d_tile (ng, iADM, Istr, Iend, Jstr, Jend,          &
     &                       LBi, UBi, LBj, UBj, 1, N(ng), Nghost,      &
     &                       DTHdif, DTVdif, Hdecay, Vdecay,            &
     &                       b_v,                                       &
     &                       pm, pn, pmon_p, pnom_r,                    &
#  ifdef MASKING
     &                       pmask,                                     &
#  endif
     &                       Hz, z_r,                                   &
     &                       ad_v(:,:,:,Linp))
!
!  Tracers.
!
      DO itrc=1,NT(ng)
        CALL ad_bcov_r3d_tile (ng, iADM, Istr, Iend, Jstr, Jend,        &
     &                         LBi, UBi, LBj, UBj, 1, N(ng), Nghost,    &
     &                         DTHdif, DTVdif, Hdecay, Vdecay,          &
     &                         b_t(:,:,:,itrc),                         &
     &                         pm, pn, pmon_u, pnom_v,                  &
# ifdef MASKING
     &                         umask, vmask,                            &
# endif
     &                         Hz, z_r,                                 &
     &                         ad_t(:,:,:,Linp,itrc))
      END DO
# endif

      RETURN
      END SUBROUTINE ad_convolution

      SUBROUTINE tl_convolution (ng, tile, Linp)
!
!=======================================================================
!                                                                      !
!  This routine computes spatial convolution of requested tangent      !
!  linear state solution.  It is used to convert from V-space  to      !
!  X-space after a call to the descent algorithm.                      !
!                                                                      !
!=======================================================================
!
      USE mod_param
# ifdef SOLVE3D
      USE mod_coupling
# endif
      USE mod_grid
      USE mod_ocean
!
!  Imported variable declarations.
!
      integer, intent(in) :: ng, tile, Linp
!
!  Local variable declarations.
!
# include "tile.h"
!
# ifdef PROFILE
      CALL wclock_on (ng, iTLM, ?)
# endif
      CALL tl_convolution_tile (ng, Istr, Iend, Jstr, Jend,             &
     &                          LBi, UBi, LBj, UBj,                     &
     &                          Linp,                                   &
     &                          GRID(ng) % pm,                          &
     &                          GRID(ng) % pmon_r,                      &
     &                          GRID(ng) % pmon_u,                      &
     &                          GRID(ng) % pn,                          &
     &                          GRID(ng) % pnom_p,                      &
     &                          GRID(ng) % pnom_v,                      &
# ifdef MASKING
     &                          GRID(ng) % pmask,                       &
     &                          GRID(ng) % rmask,                       &
     &                          GRID(ng) % umask,                       &
     &                          GRID(ng) % vmask,                       &
# endif
# ifdef SOLVE3D
     &                          GRID(ng) % h,                           &
     &                          GRID(ng) % Hz,                          &
     &                          GRID(ng) % z_r,                         &
     &                          GRID(ng) % z_w,                         &
#  ifdef ICESHELF
     &                          GRID(ng) % zice,                        &
#  endif
     &                          COUPLING(ng) % Zt_avg1,                 &
#  if defined SEdIMENT && defined SED_MORPH
     &                          OCEAN(ng) % bed,                        &
     &                          GRID(ng) % bed_thick0,                  &
#  endif
# endif
# ifdef SOLVE3D
     &                          OCEAN(ng) % b_u,                        &
     &                          OCEAN(ng) % b_v,                        &
     &                          OCEAN(ng) % b_t,                        &
# else
     &                          OCEAN(ng) % b_ubar,                     &
     &                          OCEAN(ng) % b_vbar,                     &
# endif
     &                          OCEAN(ng) % b_zeta,                     &
# ifdef SOLVE3D
     &                          OCEAN(ng) % tl_t,                       &
     &                          OCEAN(ng) % tl_u,                       &
     &                          OCEAN(ng) % tl_v,                       &
# else
     &                          OCEAN(ng) % tl_ubar,                    &
     &                          OCEAN(ng) % tl_vbar,                    &
# endif
     &                          OCEAN(ng) % tl_zeta)
# ifdef PROFILE
      CALL wclock_off (ng, iTLM, ?)
# endif
      RETURN
      END SUBROUTINE tl_convolution
!
!***********************************************************************
      SUBROUTINE tl_convolution_tile (ng, Istr, Iend, Jstr, Jend,       &
     &                                LBi, UBi, LBj, UBj,               &
     &                                Linp,                             &
     &                                pm, pmon_r, pmon_u,               &
     &                                pn, pnom_p, pnom_v,               &
# ifdef MASKING
     &                                pmask, rmask, umask, vmask,       &
# endif
# ifdef SOLVE3D
     &                                h, Hz,                            &
     &                                z_r, z_w,                         &
#  ifdef ICESHELF
     &                                zice,                             &
#  endif
     &                                Zt_avg1,                          &
#  if defined SEdIMENT && defined SED_MORPH
     &                                bed, bed_thick0,                  &
#  endif
# endif
# ifdef SOLVE3D
     &                                b_u, b_v, b_t,                    &
# else
     &                                b_ubar, b_vbar,                   &
# endif
     &                                b_zeta,                           &
# ifdef SOLVE3D
     &                                tl_u, tl_v, tl_t,                 &
# else
     &                                tl_ubar, tl_vbar,                 &
# endif
     &                                tl_zeta)
!***********************************************************************
!
      USE mod_param
      USE mod_fourdvar
      USE mod_scalars
!
      USE tl_bcov_2d_mod
# ifdef SOLVE3D
      USE tl_bcov_3d_mod
      USE set_depth_mod, ONLY : set_depth_tile
# endif
!
!  Imported variable declarations.
!
      integer, intent(in) :: ng, Iend, Istr, Jend, Jstr
      integer, intent(in) :: LBi, UBi, LBj, UBj
      integer, intent(in) :: Linp
!
# ifdef ASSUMED_SHAPE
      real(r8), intent(in) :: pm(LBi:,LBj:)
      real(r8), intent(in) :: pmon_r(LBi:,LBj:)
      real(r8), intent(in) :: pmon_u(LBi:,LBj:)
      real(r8), intent(in) :: pn(LBi:,LBj:)
      real(r8), intent(in) :: pnom_p(LBi:,LBj:)
      real(r8), intent(in) :: pnom_v(LBi:,LBj:)
#  ifdef MASKING
      real(r8), intent(in) :: pmask(LBi:,LBj:)
      real(r8), intent(in) :: rmask(LBi:,LBj:)
      real(r8), intent(in) :: umask(LBi:,LBj:)
      real(r8), intent(in) :: vmask(LBi:,LBj:)
#  endif
#  ifdef SOLVE3D
      real(r8), intent(in) :: b_u(LBi:,LBj:,:)
      real(r8), intent(in) :: b_v(LBi:,LBj:,:)
      real(r8), intent(in) :: b_t(LBi:,LBj:,:,:)
#  else
      real(r8), intent(in) :: b_ubar(LBi:,LBj:)
      real(r8), intent(in) :: b_vbar(LBi:,LBj:)
#  endif
      real(r8), intent(in) :: b_zeta(LBi:,LBj:)
#  ifdef SOLVE3D
#   ifdef ICESHELF
      real(r8), intent(in) :: zice(LBi:,LBj:)
#   endif
#   if defined SEdIMENT && defined SED_MORPH
      real(r8), intent(in) :: bed(LBi:,LBj:,:,:)

      real(r8), intent(inout) :: bed_thick0(LBi:,LBj:)
#   endif
      real(r8), intent(inout) :: h(LBi:,LBj:)
      real(r8), intent(inout) :: Hz(LBi:,LBj:,:)
      real(r8), intent(inout) :: Zt_avg1(LBi:,LBj:)
      real(r8), intent(inout) :: z_r(LBi:,LBj:,:)
      real(r8), intent(inout) :: z_w(LBi:,LBj:,:)
#  endif
#  ifdef SOLVE3D
      real(r8), intent(inout) :: tl_u(LBi:,LBj:,:,:)
      real(r8), intent(inout) :: tl_v(LBi:,LBj:,:,:)
      real(r8), intent(inout) :: tl_t(LBi:,LBj:,:,:,:)
#  else
      real(r8), intent(inout) :: tl_ubar(LBi:,LBj:,:)
      real(r8), intent(inout) :: tl_vbar(LBi:,LBj:,:)
#  endif
      real(r8), intent(inout) :: tl_zeta(LBi:,LBj:,:)

# else

      real(r8), intent(in) :: pm(LBi:UBi,LBj:UBj)
      real(r8), intent(in) :: pmon_r(LBi:UBi,LBj:UBj)
      real(r8), intent(in) :: pmon_u(LBi:UBi,LBj:UBj)
      real(r8), intent(in) :: pn(LBi:UBi,LBj:UBj)
      real(r8), intent(in) :: pnom_p(LBi:UBi,LBj:UBj)
      real(r8), intent(in) :: pnom_v(LBi:UBi,LBj:UBj)
#  ifdef MASKING
      real(r8), intent(in) :: pmask(LBi:UBi,LBj:UBj)
      real(r8), intent(in) :: rmask(LBi:UBi,LBj:UBj)
      real(r8), intent(in) :: umask(LBi:UBi,LBj:UBj)
      real(r8), intent(in) :: vmask(LBi:UBi,LBj:UBj)
#  endif
#  ifdef SOLVE3D
      real(r8), intent(in) :: b_u(LBi:UBi,LBj:UBj,N(ng))
      real(r8), intent(in) :: b_v(LBi:UBi,LBj:UBj,N(ng))
      real(r8), intent(in) :: b_t(LBi:UBi,LBj:UBj,N(ng),NT(ng))
#  else
      real(r8), intent(in) :: b_ubar(LBi:UBi,LBj:UBj)
      real(r8), intent(in) :: b_vbar(LBi:UBi,LBj:UBj)
#  endif
      real(r8), intent(in) :: b_zeta(LBi:UBi,LBj:UBj)
#  ifdef SOLVE3D
#   ifdef ICESHELF
      real(r8), intent(in) :: zice(LBi:UBi,LBj:UBj)
#   endif
#   if defined SEdIMENT && defined SED_MORPH
      real(r8), intent(in) :: bed(LBi:UBi,LBj:UBj,Nbed,MBEDP)

      real(r8), intent(inout) :: bed_thick0(LBi:UBi,LBj:UBj)
#   endif
      real(r8), intent(inout) :: h(LBi:UBi,LBj:UBj)
      real(r8), intent(inout) :: Hz(LBi:UBi,LBj:UBj,N(ng))
      real(r8), intent(inout) :: Zt_avg1(LBi:UBi,LBj:UBj)
      real(r8), intent(inout) :: z_r(LBi:UBi,LBj:UBj,N(ng))
      real(r8), intent(inout) :: z_w(LBi:UBi,LBj:UBj,0:N(ng))
#  endif
#  ifdef SOLVE3D
      real(r8), intent(inout) :: tl_u(LBi:UBi,LBj:UBj,N(ng),2)
      real(r8), intent(inout) :: tl_v(LBi:UBi,LBj:UBj,N(ng),2)
      real(r8), intent(inout) :: tl_t(LBi:UBi,LBj:UBj,N(ng),3,NT(ng))
#  else
      real(r8), intent(inout) :: tl_ubar(LBi:UBi,LBj:UBj,3)
      real(r8), intent(inout) :: tl_vbar(LBi:UBi,LBj:UBj,3)
#  endif
      real(r8), intent(inout) :: tl_zeta(LBi:UBi,LBj:UBj,3)
# endif
!
!  Local variable declarations.
!
      integer :: IstrR, IendR, JstrR, JendR, IstrU, JstrV
      integer :: i, itrc, j
!
# include "set_bounds.h"

# ifdef SOLVE3D
!
!-----------------------------------------------------------------------
!  Compute depths and thicknesses for convolution. Use a rest state
!  free-surface.
!-----------------------------------------------------------------------
!
      DO j=JstrR,JendR
        DO i=IstrR,IendR
          Zt_avg1(i,j)=0.0_r8
        END DO
      END DO
      CALL set_depth_tile (ng, Istr, Iend, Jstr, Jend,                  &
     &                     LBi, UBi, LBj, UBj,                          &
     &                     h,                                           &
#  ifdef ICESHELF
     &                     zice,                                        &
#  endif
#  if defined SEDIMENT && defined SED_MORPH
     &                     bed, bed_thick0,                             &
#  endif
     &                     Zt_avg1,                                     &
     &                     Hz, z_r, z_w)
# endif
!
!-----------------------------------------------------------------------
!  Adjoint spatial convolutions.
!-----------------------------------------------------------------------
!
!  Free-surface.
!
      CALL tl_bcov_r2d_tile (ng, iADM, Istr, Iend, Jstr, Jend,          &
     &                       LBi, UBi, LBj, UBj, Nghost,                &
     &                       DTHdif, Hdecay,                            &
     &                       b_zeta,                                    &
     &                       pm, pn, pmon_u, pnom_v,                    &
# ifdef MASKING
     &                       umask, vmask,                              &
# endif
     &                       tl_zeta(:,:,Linp))

# ifndef SOLVE3D
!
!  2D U-velocity.
!
      CALL tl_bcov_u2d_tile (ng, iADM, Istr, Iend, Jstr, Jend,          &
     &                       LBi, UBi, LBj, UBj, Nghost,                &
     &                       DTHdif, Hdecay,                            &
     &                       b_ubar,                                    &
     &                       pm, pn, pmon_r, pnom_p,                    &
#  ifdef MASKING
     &                       pmask,                                     &
#  endif
     &                       tl_ubar(:,:,Linp))
!
!  2D V-velocity.
!
      CALL tl_bcov_v2d_tile (ng, iADM, Istr, Iend, Jstr, Jend,          &
     &                       LBi, UBi, LBj, UBj, Nghost,                &
     &                       DTHdif, Hdecay,                            &
     &                       b_vbar,                                    &
     &                       pm, pn, pmon_p, pnom_r,                    &
#  ifdef MASKING
     &                       pmask,                                     &
#  endif
     &                       tl_vbar(:,:,Linp))
# else
!
!  3D U-velocity.
!
      CALL tl_bcov_u3d_tile (ng, iADM, Istr, Iend, Jstr, Jend,          &
     &                       LBi, UBi, LBj, UBj, 1, N(ng), Nghost,      &
     &                       DTHdif, DTVdif, Hdecay, Vdecay,            &
     &                       b_u,                                       &
     &                       pm, pn, pmon_r, pnom_p,                    &
#  ifdef MASKING
     &                       pmask,                                     &
#  endif
     &                       Hz, z_r,                                   &
     &                       tl_u(:,:,:,Linp))
!
!  3D V-velocity.
!
      CALL tl_bcov_v3d_tile (ng, iADM, Istr, Iend, Jstr, Jend,          &
     &                       LBi, UBi, LBj, UBj, 1, N(ng), Nghost,      &
     &                       DTHdif, DTVdif, Hdecay, Vdecay,            &
     &                       b_v,                                       &
     &                       pm, pn, pmon_p, pnom_r,                    &
#  ifdef MASKING
     &                       pmask,                                     &
#  endif
     &                       Hz, z_r,                                   &
     &                       tl_v(:,:,:,Linp))
!
!  Tracers.
!
      DO itrc=1,NT(ng)
        CALL tl_bcov_r3d_tile (ng, iADM, Istr, Iend, Jstr, Jend,        &
     &                         LBi, UBi, LBj, UBj, 1, N(ng), Nghost,    &
     &                         DTHdif, DTVdif, Hdecay, Vdecay,          &
     &                         b_t(:,:,:,itrc),                         &
     &                         pm, pn, pmon_u, pnom_v,                  &
# ifdef MASKING
     &                         umask, vmask,                            &
# endif
     &                         Hz, z_r,                                 &
     &                         tl_t(:,:,:,Linp,itrc))
      END DO
# endif

      RETURN
      END SUBROUTINE tl_convolution
#endif
      END MODULE convolution_mod
