#include "cppdefs.h"
#if defined IMPULSE && defined SOLVE3D
      SUBROUTINE get_3dfldi (ng, model, ifield, ncid, nfiles, fname,    &
     &                       update, LBi, UBi, LBj, UBj, LBk, UBk,      &
     &                       Iout, Irec,                                &
# ifdef MASKING
     &                       Fmask,                                     &
# endif
     &                       Fout)
!
!=======================================================================
!  Copyright (c) 2005 ROMS/TOMS Group                                  !
!================================================== Hernan G. Arango ===
!                                                                      !
!  This routine reads in requested 3D field from specified NetCDF      !
!  file. Forward time and adjoint impulse forcing processing.          !
!                                                                      !
!  On Input:                                                           !
!                                                                      !
!     ng         Nested grid number.                                   !
!     model      Calling model identifier.                             !
!     ifield     Field ID.                                             !
!     ncid       NetCDF file ID.                                       !
!     nfiles     Number of input NetCDF files.                         !
!     fname      NetCDF file name(s).                                  !
!     LBi        I-dimension Lower bound.                              !
!     UBi        I-dimension Upper bound.                              !
!     LBj        J-dimension Lower bound.                              !
!     UBj        J-dimension Upper bound.                              !
!     LBk        K-dimension Lower bound.                              !
!     UBk        K-dimension Upper bound.                              !
!     Iout       Size of the outer dimension,  if any.  Otherwise,     !
!                  Iout must be set to one by the calling program.     !
!     Irec       Number of 3D field records to read.                   !
!     Fmask      Land/Sea mask, if any.                                !
!                                                                      !
!  On Output:                                                          !
!                                                                      !
!     Fout       Read field.                                           !
!     update     Switch indicating reading of the requested field      !
!                  the current time step.                              !
!                                                                      !
!=======================================================================
!
      USE mod_param
      USE mod_parallel
      USE mod_iounits
      USE mod_ncparam
      USE mod_netcdf
      USE mod_scalars

      implicit none
!
!  Imported variable declarations.
!
      logical, intent(out) :: update

      integer, intent(in) :: ng, model, ifield, ncid, nfiles
      integer, intent(in) :: LBi, UBi, LBj, UBj, LBk, UBk
      integer, intent(in) :: Iout, Irec

      character (len=*), intent(in) :: fname(nfiles)

# ifdef MASKING
      real(r8), intent(in) :: Fmask(LBi:UBi,LBj:UBj)
# endif
      real(r8), intent(inout) :: Fout(LBi:UBi,LBj:UBj,LBk:UBk,Iout)
!
!  Local variable declarations.
!
      logical :: Liocycle, Lgrided, Lonerec, got_var, got_time

      integer :: Tid, Tindex, Trec, Vid, Vtype
      integer :: i, lend, lfvar, lstr, ltvar, lvar, nrec, nvatts
      integer :: status, vartype

      integer :: nf_fread

      real(r8) :: Clength, Fmax, Fmin, Fval, Tdelta, Tend
      real(r8) :: Tmax, Tmin, Tmono, Tscale, Tstr, Tval, scale

      character (len=45) :: attnam
      character (len=80) :: ncfile
!
!-----------------------------------------------------------------------
!  On first call, inquire about the contents of input NetCDF file.
!-----------------------------------------------------------------------
!
      IF (exit_flag.ne.NoError) RETURN
!
      IF (iic(ng).eq.0) THEN
!
!  Intialize local variables.
!
        Vid=-1
        Tid=-1        
        Liocycle=.FALSE.
        Lgrided=.FALSE.
        Lonerec=.FALSE.
        got_var=.FALSE.
        got_time=.FALSE.
        Vtype=Iinfo(1,ifield,ng)
        lfvar=LEN_TRIM(Vname(1,ifield))
        ltvar=LEN_TRIM(Tname(ifield))
!
!  Inquire about the dimensions and variables. Check for consistency.
!
        IF (InpThread) THEN
          CALL opencdf (ng, nfiles, fname, ncfile, N(ng), ifield, nrec)
          IF (exit_flag.ne.NoError) RETURN
!
!  Scan variable list from input NetCDF and check for requested
!  variables.
!
          DO i=1,nvars
            IF (lfvar.gt.0) THEN
              IF (TRIM(varnam(i)).eq.TRIM(Vname(1,ifield))) THEN
                Vid=i
                got_var=.TRUE.
                IF (nvdims(i).gt.1) Lgrided=.TRUE.
              END IF
            END IF
            IF (ltvar.gt.0) THEN
              IF (TRIM(varnam(i)).eq.TRIM(Tname(ifield))) THEN
                Tid=i
                got_time=.TRUE.
              END IF
            END IF
          END DO
          LinfoAD(1,ifield,ng)=Lgrided
          IinfoAD(1,ifield,ng)=Vtype
          IinfoAD(2,ifield,ng)=Vid
          IinfoAD(3,ifield,ng)=Tid
          IinfoAD(4,ifield,ng)=nrec
!
!  Terminate execution requested variables are not found.
!
          IF (.not.got_var.and.(lfvar.gt.0)) THEN
            WRITE (stdout,10) TRIM(Vname(1,ifield)), TRIM(ncfile)
            exit_flag=2
            RETURN
          END IF
          IF (.not.got_time.and.(ltvar.gt.0)) THEN
            WRITE (stdout,10) TRIM(Tname(ifield)), TRIM(ncfile)
            exit_flag=2
            RETURN
          END IF
!
!  If appropriate, open input NetCDF file for reading.  If processing
!  model forcing (multiple files allowed), check if file for requested
!  field has been already opened and get/save its ID from/to the
!  association table. 
!
          CALL openids (nFfiles(ng), FRCname(1,ng), FRCids(1,ng),       &
     &                  ncfile, ncid)
          IF (ncid.eq.-1) THEN
            status=nf_open(TRIM(ncfile), nf_nowrite, ncid)
            IF (status.ne.nf_noerr) THEN
              WRITE (stdout,20) TRIM(ncfile)
              exit_flag=2
              ioerror=status
              RETURN
            END IF
            CALL openids (nFfiles(ng), FRCname(1,ng), FRCids(1,ng),     &
     &                    ncfile, ncid)
          END IF
          CinfoAD(ifield,ng)=TRIM(ncfile)
!
!  If "scale_factor" attribute is present for a variable, the data are
!  to be multiplied by this factor.  Check if only water points are
!  available.
!
          IF (got_var) THEN
            status=nf_inq_var(ncid, Vid, varnam(Vid), vartype,          &
     &                        nvdims(Vid), vdims(1,Vid), nvatts)
            IF (status.eq.nf_noerr) THEN
              DO i=1,nvatts
                status=nf_inq_attname(ncid, Vid, i, attnam)
                IF (status.eq.nf_noerr) THEN
                  IF (TRIM(attnam).eq.'scale_factor') THEN
                    status=nf_get_att_TYPE(ncid, Vid, TRIM(attnam),     &
     &                                     scale)
                    IF (status.eq.nf_noerr) THEN
                      FscaleAD(ifield,ng)=FscaleAD(ifield,ng)*scale
                    ELSE
                      WRITE (stdout,30) TRIM(attnam)
                      exit_flag=2
                      RETURN
                    END IF
                  ELSE IF (TRIM(attnam).eq.'water_points') THEN
                    IinfoAD(1,ifield,ng)=-ABS(IinfoAD(1,ifield,ng))
                    Vtype=IinfoAD(1,ifield,ng)
                  END IF
                ELSE
                  WRITE (stdout,40) TRIM(varnam(Vid))
                  exit_flag=2
                  RETURN
                END IF
              END DO
            ELSE
              WRITE (stdout,50)
              exit_flag=2
              RETURN
            END IF
          END IF
!
!  Determine initial time record to read and cycling switch.
!
          CALL get_cycle (ng, ncid, Tid, nrec, tdays(ng), Liocycle,     &
     &                    Clength, Trec, Tstr, Tend, Tmin, Tmax, Tscale)
          IF (exit_flag.ne.NoError) RETURN
          LinfoAD(2,ifield,ng)=Liocycle
          FinfoAD(1,ifield,ng)=Tmin
          FinfoAD(2,ifield,ng)=Tmax
          FinfoAD(3,ifield,ng)=Tstr
          FinfoAD(4,ifield,ng)=Tend
          FinfoAD(5,ifield,ng)=Clength
          FinfoAD(6,ifield,ng)=Tscale
!
!  The strategy here is to create a local, monotonically increasing
!  time variable so the interpolation between snapshots is trivial
!  when cycling forcing fields. Subtract one to time record counter
!  "Trec" to avoid doing special case at initialization.
!
          IF (Irec.eq.1) THEN
            Tindex=2
          ELSE
            Tindex=1
          END IF
          IF (Liocycle) THEN
            IF (Trec.eq.nrec) THEN
              IF (tdays(ng).lt.Tmax) THEN
                Tmono=Tstr-Clength
              ELSE
                Tmono=tdays(ng)+(Tstr-Clength)
                IF (Tstr.eq.Tmax) THEN
                  Tmono=Tmono+(Tmin-MOD(tdays(ng)+Tmin,Clength))
                ELSE
                  Tmono=Tmono+(Tstr-MOD(tdays(ng)+Tstr,Clength))
                END IF
              END IF
            ELSE
              IF (tdays(ng).gt.Clength) THEN
                Tmono=tdays(ng)-MOD(tdays(ng)-Tstr,Clength)
              ELSE
                Tmono=Tstr
              END IF
            END IF
          ELSE
            Tmono=Tstr
          END IF
          Tmono=Tmono*day2sec
          Trec=Trec-1
          IinfoAD(5,ifield,ng)=Tindex
          IinfoAD(6,ifield,ng)=Trec
          FinfoAD(7,ifield,ng)=Tmono
!
!  Set switch for one time record dataset. In this case, the input
!  data is always the same and time interpolation is not performed.
!
          IF (nrec.eq.1) Lonerec=.TRUE.
          LinfoAD(3,ifield,ng)=Lonerec
        END IF
# ifdef DISTRIBUTE
!
!  Broadcast variable information to all nodes in the group.
!
        CALL mp_bcastl (ng, model, LinfoAD(1,ifield,ng), 3)
        CALL mp_bcasti (ng, model, IinfoAD(1,ifield,ng), 6)
        CALL mp_bcastf (ng, model, FinfoAD(1,ifield,ng), 7)
        CALL mp_bcasts (ng, model, CinfoAD(ifield,ng), 80)
# endif
        Tindex=IinfoAD(5,ifield,ng)
        VtimeAD(Tindex,ifield,ng)=FinfoAD(3,ifield,ng)
      END IF         
!
!-----------------------------------------------------------------------
!  Get requested field information from global storage.
!-----------------------------------------------------------------------
!
      Lgrided =LinfoAD(1,ifield,ng)
      Liocycle=LinfoAD(2,ifield,ng)
      Lonerec =LinfoAD(3,ifield,ng)
      Vtype   =IinfoAD(1,ifield,ng)
      Vid     =IinfoAD(2,ifield,ng)
      Tid     =IinfoAD(3,ifield,ng)
      nrec    =IinfoAD(4,ifield,ng)
      Tindex  =IinfoAD(5,ifield,ng)
      Trec    =IinfoAD(6,ifield,ng)
      Clength =FinfoAD(5,ifield,ng)
      Tscale  =FinfoAD(6,ifield,ng)
      Tmono   =FinfoAD(7,ifield,ng)
      ncfile  =CinfoAD(ifield,ng)
!
!-----------------------------------------------------------------------
!  If appropriate, read in new data.
!-----------------------------------------------------------------------
!
      update=.FALSE.
      IF ((Tmono.le.time(ng)).or.(iic(ng).eq.0).or.                     &
     &    (iic(ng).eq.ntstart)) THEN
        IF (Liocycle) THEN
          Trec=MOD(Trec,nrec)+1
        ELSE
          Trec=Trec+1
        END IF
        IinfoAD(6,ifield,ng)=Trec
        IF (Trec.le.nrec) THEN
!
!  Set rolling index for two-time record storage of input data.  If
!  "Iout" is unity, input data is stored in timeless array by the
!  calling program.  If Irec > 1, this routine is used to read a 3D
!  field varying in another non-time dimension.
!
          IF (Irec.eq.1) THEN
            IF (Iout.eq.1) THEN
              Tindex=1
            ELSE
              Tindex=3-Tindex
            END IF
            IinfoAD(5,ifield,ng)=Tindex
          END IF
!
!  Read in time coordinate and scale it to day units.
!
          IF (InpThread.and.(Tid.ge.0)) THEN
            status=nf_get_var1_TYPE(ncid, Tid, Trec, Tval)
            IF (status.ne.nf_noerr) THEN
              WRITE (stdout,60) TRIM(Tname(ifield)), Trec
              exit_flag=2
              ioerror=status
              RETURN
            END IF
          END IF
# ifdef DISTRIBUTE
          CALL mp_bcastf (ng, model, Tval, 1)
# endif
          VtimeAD(Tindex,ifield,ng)=Tval*Tscale
!
!  Read in 3D-grided or point data.
!
          IF (Vid.ge.0) THEN
            Fmin=0.0_r8
            Fmax=0.0_r8
            IF (Lgrided) THEN
              IF (Irec.gt.1) THEN
                FinfoAD(8,ifield,ng)=1.0E+35_r8
                FinfoAD(9,ifield,ng)=-1.0E+35_r8
                DO i=1,Irec
                  status=nf_fread(ng, model, ncid, Vid, i, Vtype,       &
     &                            LBi, UBi, LBj, UBj, LBk, UBk,         &
     &                            FscaleAD(ifield,ng), Fmin, Fmax,      &
# ifdef MASKING
     &                            Fmask(LBi,LBj),                       &
# endif
     &                            Fout(LBi,LBj,LBk,i))
                  FinfoAD(8,ifield,ng)=MIN(Fmin,FinfoAD(8,ifield,ng))
                  FinfoAD(9,ifield,ng)=MAX(Fmax,FinfoAD(9,ifield,ng))
                END DO
              ELSE
                status=nf_fread(ng, model, ncid, Vid, Trec, Vtype,      &
     &                          LBi, UBi, LBj, UBj, LBk, UBk,           &
     &                          FscaleAD(ifield,ng), Fmin, Fmax,        &
# ifdef MASKING
     &                          Fmask(LBi,LBj),                         &
# endif
     &                          Fout(LBi,LBj,LBk,Tindex))
                FinfoAD(8,ifield,ng)=Fmin
                FinfoAD(9,ifield,ng)=Fmax
              END IF
            END IF
            IF (InpThread.and.(status.ne.nf_noerr)) THEN
              WRITE (stdout,60) TRIM(Vname(1,ifield)), Trec
              exit_flag=2
              ioerror=status
              RETURN
            END IF
            IF (Master) THEN
              IF (Irec.gt.1) THEN
                WRITE (stdout,70) TRIM(Vname(2,ifield)), Fmin, Fmax
              ELSE
                lstr=SCAN(ncfile,'/',BACK=.TRUE.)+1
                lend=LEN_TRIM(ncfile)
                lvar=MIN(43,LEN_TRIM(Vname(2,ifield)))
                WRITE (stdout,80) Vname(2,ifield)(1:lvar),              &
     &                            Tval*Tscale, ncfile(lstr:lend),       &
     &                            Trec, Tindex, Fmin, Fmax
              END IF
            END IF
            update=.TRUE.
          END IF
        END IF
!
!  Increment the local time variable "Tmono" by the interval between
!  snapshots. If the interval is negative, indicating cycling, add in
!  a cycle length.  Load time value (sec) into "TintrpAD" which used
!  during interpolation between snapshots.
!
        IF (.not.Lonerec) THEN
          Tdelta=VtimeAD(Tindex,ifield,ng)-VtimeAD(3-Tindex,ifield,ng)
          IF (Liocycle.and.(Tdelta.lt.0.0_r8)) THEN
            Tdelta=Tdelta+Clength
          END IF
          Tmono=Tmono+Tdelta*day2sec
          FinfoAD(7,ifield,ng)=Tmono
          TintrpAD(Tindex,ifield,ng)=Tmono
        END IF
      END IF
!
  10  FORMAT (/,' GET_3DFLDI  - unable to find requested variable: ',a, &
     &        /,15x,'in input NetCDF file: ',a)
  20  FORMAT (/,' GET_3DFLDI  - unable to open input NetCDF file: ',a)
  30  FORMAT (/,' GET_3DFLDI  - error while reading attribute: ',a)
  40  FORMAT (/,' GET_3DFLDI  - error while inquiring attributes for',  &
     &        ' variable: ',a)
  50  FORMAT (/,' GET_3DFLDI  - cannot inquire about time variable',    &
     &          ' from input NetCDF file.')
  60  FORMAT (/,' GET_3DFLDI  - error while reading variable: ',a,2x,   &
     &       ' at TIME index = ',i4)
  70  FORMAT (3x,' GET_3DFLDI  - ',a,/,19x,'(Min = ',1p,e15.8,0p,       &
     &        ' Max = ',1p,e15.8,0p,')')
  80  FORMAT (3x,' GET_3DFLDI  - ',a,',',t64,'t = ',f12.4,/,19x,        &
     &        '(File: ',a,', Rec=',i4.4,', Index=',i1,')',/,19x,        &
     &        '(Min = ', 1p,e15.8,0p,' Max = ',1p,e15.8,0p,')')
      RETURN
      END SUBROUTINE get_3dfldi
#else
      SUBROUTINE get_3dfldi
      RETURN
      END SUBROUTINE get_3dfldi
#endif
