#include "cppdefs.h"
#ifdef FOUR_DVAR
      SUBROUTINE obs_cost (ng, model)
!
!=================================================== Andrew M. Moore ===
!  Copyright (c) 2005 ROMS/TOMS Adjoint Group                          !
!================================================== Hernan G. Arango ===
!                                                                      !
!  This routine computes the observation cost function (Jo) as the     !
!  misfit (squared difference) between model and observations:         !
!                                                                      !
!         Jo = transpose(X - Xo) * O^(-1) * (X -Xo)                    !
!                                                                      !
!  where                                                               !
!                                                                      !
!         Xo : observations                                            !
!         X  : model at observation points                             !
!         O  : observations error covariance (assigned weight)         !
!                                                                      !
!=======================================================================
!
      USE mod_param
      USE mod_parallel
      USE mod_fourdvar
!
      implicit none
!
!  Imported variable declarations.
!
      integer, intent(in) :: ng, model
!
!  Local variable declarations.
!
      integer :: NSUB, iobs, ivar

      real(r8) ::  cff, my_StepSizeB, my_StepSizeT

      real(r8), dimension(0:NstateVar(ng)) :: my_cost

# ifdef DISTRIBUTE
      real(r8), dimension(0:NstateVar(ng)+2) :: buffer
      character (len=3), dimension(0:NstateVar(ng)+2) :: op_handle
# endif
!
!-----------------------------------------------------------------------
!  Compute observation norm scale (ObsNorm).
!-----------------------------------------------------------------------
!
      CALL obs_norm (ng, model)
!
!-----------------------------------------------------------------------
!  Compute observation misfit cost function (CostFun).
!-----------------------------------------------------------------------
!
      DO ivar=0,NstateVar(ng)
        my_cost(ivar)=0.0_r8
      END DO
      IF (Ipass.eq.1) THEN
        DO iobs=1,Nobs(ng)
          ivar=ObsType(iobs)
# if defined S4DVAR
          cff=ObsNorm(iobs)*ObsErr(iobs)*                               &
     &        (NLmodVal(iobs,Ipass)-ObsVal(iobs))**2
# elif defined TLM_OBS
          cff=ObsNorm(iobs)*ObsErr(iobs)*                               &
     &        (NLmodVal(iobs,1)+TLmodVal(iobs,Ipass)-ObsVal(iobs))**2
# endif
          my_cost(0)=my_cost(0)+cff
          my_cost(ivar)=my_cost(ivar)+cff
        END DO
      END IF
# if defined S4DVAR || defined IS4DVAR
!
!-----------------------------------------------------------------------
!  Compute Conjugate gradient algorithm step size dot products,
!  StepSizeT, StepSizeB:
!
!               CGstepR=CGstepF*(StepSizeT/StepSizeB)
!-----------------------------------------------------------------------
!
      IF (Ipass.eq.2) THEN
        my_StepSizeT=0.0_r8
        my_StepSizeB=0.0_r8
        DO iobs=1,Nobs(ng)
          cff=ObsNorm(iobs)*ObsErr(iobs)
#  if defined S4DVAR
          my_StepSizeT=my_StepSizeT-                                    &
     &                 cff*                                             &
     &                 (NLmodVal(iobs,Ipass-1)-ObsVal(iobs))*           &
     &                 (NLmodVal(iobs,Ipass)-NLmodVal(iobs,Ipass-1))
          my_StepSizeB=my_StepSizeB+                                    &
     &                 cff*                                             &
     &                 (NLmodVal(iobs,Ipass)-NLmodVal(iobs,Ipass-1))**2
#  elif defined IS4DVAR
          my_StepSizeT=my_StepSizeT-                                    &
     &                 cff*                                             &
     &                 (NLmodVal(iobs,1)+TLmodVal(iobs,Ipass-1)-        &
     &                  ObsVal(iobs))*                                  &
     &                 (TLmodVal(iobs,Ipass)-TLmodVal(iobs,Ipass-1))
          my_StepSizeB=my_StepSizeB+                                    &
     &                 cff*                                             &
     &                 (TLmodVal(iobs,Ipass)-TLmodVal(iobs,Ipass-1))**2
#  endif 
        END DO
      END IF
# endif
!
!-----------------------------------------------------------------------
!  Compute global values.
!-----------------------------------------------------------------------
!
      NSUB=NtileX(ng)*NtileE(ng)
!$OMP CRITICAL (COST_FUN)
      IF (tile_count.eq.0) THEN
        DO ivar=0,NstateVar(ng)
          FOURDVAR(ng)%CostFun(ivar)=my_cost(ivar)
        END DO
        StepSizeB=my_StepSizeB
        StepSizeT=my_StepSizeT
      ELSE
        DO ivar=0,NstateVar(ng)
          FOURDVAR(ng)%CostFun(ivar)=FOURDVAR(ng)%CostFun(ivar)+        &
     &                               my_cost(ivar)
        END DO
        StepSizeB=StepSizeB+my_StepSizeB
        StepSizeT=StepSizeT+my_StepSizeT
      END IF
      tile_count=tile_count+1
      IF (tile_count.eq.NSUB) THEN
        tile_count=0
#ifdef DISTRIBUTE
        DO ivar=0,NstateVar(ng)
          buffer(ivar)=FOURDVAR(ng)%CostFun(ivar)
          op_handle(ivar)='SUM'
        END DO
        buffer(NstateVar(ng)+1)=StepSizeB
        buffer(NstateVar(ng)+2)=StepSizeT
        op_handle(NstateVar(ng)+1)='SUM'
        op_handle(NstateVar(ng)+2)='SUM'
        CALL mp_reduce (ng, iNLM, NstateVar(ng)+3, buffer(0:),          &
     &                  op_handle(0:))
        DO ivar=0,NstateVar(ng)
          FOURDVAR(ng)%CostFun(ivar)=buffer(ivar)
        END DO
        StepSizeB=buffer(NstateVar(ng)+1)
        StepSizeT=buffer(NstateVar(ng)+2)
#endif
      END IF
!$OMP END CRITICAL (COST_FUN)

      RETURN
      END SUBROUTINE obs_cost
#else
      SUBROUTINE obs_cost
      RETURN
      END SUBROUTINE obs_cost
#endif
